//The following must be executed, to 
//global N O M;

//Finds an optimal dynamic schedule, given the set of training instances
//for bound optimization, the bound is equivalent to the end time
//icount - number of intervals, including the last one
//The end time = icount - 1
//t - type of optimization, t = 1 mean optimization, t = 2 limit optimization
//Example: dynOptimize(X,9,2)
function R = dynOptimize(X, icount, t)
	//N - new state values
	//O - old state values, they are replaced by new to save memory
	//M - successions, define which algorithm is optimal in which state
	global N O M;

	m = size(X, 2);
	s = ones(m,1)*icount;
	[p, q] = stateToPosition(s,icount);
	N = spzeros(p, q);
	M = spzeros(p, q);

	generateStates(zeros(m,1), icount-1, 1, icount, 0, t);
	O = N;
	for i=(icount-2):-1:0
		N = spzeros(p, q);
		generateStates(zeros(m,1), i, 1, icount, 1, t);
		O = N;
	end
	R = O;
endfunction


//Reconstructs the dynamic schedule from results calculated by
//dynOptimize
//algs - number of algorithms
//icount - number of intervals
function V = reconstructSequence(algs, icount)
	global M;
	s = zeros(algs, 1);
	V = zeros(icount);
	for i=1:icount do
		[p, q] = stateToPosition(s, icount);
		k = full(M(p,q));
		V(i) = k;
		s(k) = s(k)+1;
	end
endfunction

//Simulates a dynamic schedule on the provided instances
//X - instances
//V - dynamic schedule
//t - type, t = 1 mean, t = 2 bound simulation
function R = simulateD(X,V)
	R = zeros(size(X,1),1);
	S = zeros(size(X,1),1);
	s = zeros(size(X,2),1);
	for i=1:(size(V,1)-1) do
		k = V(i);
		H = X(:,k) >= s(k) & X(:,k) < s(k) + 1
		H = H & ~S;
		R(H) = X(H,k) + sum(s) - s(k);
		S = S | H;
		s(k) = s(k) + 1;
	end
	//similar, but for the last interval
	k = V($);
	H = X(:,k) >= s(k);
	H = H & ~S;
	R(H) = X(H,k) + sum(s) - s(k);
endfunction

//Calculates the reward of the last final states
//t - type of optimization, 
//t = 1 mean optimization, t = 2 bound optimization
//s - the state
//k - the algorithm that was run
//X - the algorithm instance performance matrix
function r = calculateEndStateReward(s, k, X, t)
	C = zeros(size(X,1), 1);
	for j=1:size(X,1) do
		C(j) = sum((X(j,:) - s') < 0) == 0;
	end
	C = bool2s(C);
	if sum(C) > 0 then
		r = sum(X(:,k).*C) / sum(C) + sum(s)	- s(k);
	else
		r = 0;
	end
endfunction

//Calculates the reward and probability for solving an instance
//in the states that precede the last state
//k - the algorithm to be run
function [p,r] = calculateInnerStateReward(s, k, X, t)
	C = zeros(size(X,1), 1);
	D = zeros(size(X,1), 1);
	h = s(k)+1;
	
	for j=1:size(X,1) do
		C(j) = (sum(X(j,:) < s') == 0);
		D(j) = (X(j,k) < h);
	end
	
	C = bool2s(C);
	D = bool2s(D);
	E = C .* D;
	
	if(sum(E) > 0)
		if t == 1 then
			r = sum(X(:,k).*E)/sum(E) + sum(s)	- s(k);
		else
			r = -1;
		end;
		p = sum(E) / sum(C);	
	else
		r = 0;
		p = 0;
	end
endfunction

//Calculate the utility of inner states from S, 
//that is those that do not lead to last states is F
//s - the state
function calculateInnerStateUtility(s, icount, t)
	global N O M;
	
	[p,q] = stateToPosition(s,icount);		
	if t == 1 then
		start = 1;
	else
		start = lastNonZero(s);
	end

	for k=start:size(s,1) do
		h = s; h(k) = h(k) + 1;		
		[p2,q2] = stateToPosition(h,icount);
		[pp, pr] = calculateInnerStateReward(s, k, X, t);
		r = pp*pr + full(O(p2,q2))*(1-pp);
		
		o = full(N(p,q));
		if o == 0 then
			N(p,q) = r;
			M(p,q) = k;
		else
			if r < o then
				N(p,q) = r;
				M(p,q) = k;
			end
		end
	end
endfunction

//Calculate the utility of end states from S, 
//that is those that lead to last states is F
//s - the state
function calculateEndStateUtility(s, icount, t)
	global N M;
	[p,q] = stateToPosition(s,icount);		
	for k=1:size(s,1) do
		r = calculateEndStateReward(s, k, X, t)
		o = full(N(p,q));
		if o == 0 then
			N(p,q) = r;
			M(p,q) = k;
		else
			if r < o then
				N(p,q) = r;
				M(p,q) = k;
			end
		end
	end
endfunction

//Generates successively states and calculates their utility
//s - the current state
//l - the starting point
//g - size total sum
//O - old values, sparse matrix
//X - the statistics
//icount - the maximal value allowed for each state
//N - new values, sparse matrix
//ie == 0 -> end state, otherwise inner state
//Example: generateStates(zeros(3,1), 9, 1, 9, 0)
function generateStates(s, g, l, icount, ie, t)
	global N O;	
	if l >= size(s,1) then
		s(l) = g;
		if ie == 0 then
			if t == 1 then	
				calculateEndStateUtility(s, icount, t);
			end		
		else
			calculateInnerStateUtility(s, icount, t);
		end
	else
		for i=0:g do
			s(l) = i;
			generateStates(s, g-i, l+1, icount, ie, t);
		end
	end
endfunction



//************ helper functions

//Transforms a state into coordinates in a 2-dimensional matrix
//s - state vector - column
//m - the maximal value for each position
//p - the first dimension
//q - the second dimension
function [p,q] = stateToPosition(s, m)
	p = 0;
	q = 0;

	k = 1;
	l = size(s, 1);

	for i=floor(l/2):-1:1 do
		p = p + k * s(i);	
		k = k*m;
	end
	k=1;
	for i=l:-1:(floor(l/2)+1) do
		q = q + k * s(i);
		k = k*m;
	end
	p=p+1;	q=q+1;
endfunction

//returns the last non-zero element of the vector,
//returns 1 for a zero vector
function i=lastNonZero(s)
	for i=size(s,1):-1:1 do
		if s(i) ~= 0 then
			break;
		end
	end
endfunction

