//X contains profits, which are equivalent to negative solution times
//The following code loads the test instances
//X = fscanfMat("<path>");
//X = modifyEndValue(X, -20000);
//X1 = X(1:400,:);
//X2 = X(400:800, :);
//X3 = X(800:1200, :);
//X4 = X(503:1303, :);
//X5 = X(1:800, :);
//X6 = X(1100:1302, :);
//X7 = X(700:900, :);
//Xx = eliminateUnsolvable(X);

//X6A1 = X6(:,[3,19,21]);
//X6A2 = X6(:,[7,20,23]);
//X6A3 = X6(:,[5,6,8]);
//X7A1 = X7(:,[3,19,21]);
//X7A2 = X7(:,[7,20,23]);
//X7A3 = X7(:,[5,6,8]);

//Generally, the semantics of the symbols is following:
//X - performance of algorithms; instances x algorithms
//W - assignment of solutions to algorithms, has the same dimensions as X
//c - schedule - vector of r values

//Simulates the run of the parallel portfolio with a static schedule
function z = simulate(X,c)
	ieee(2);
	W = calculateW(X,c);
	r = ones(size(c,1),1)./c;
	r = naneliminate(r); 
	//This will not work for positive profits and min, because then
	//zero will be minimal value.
	z = X .* W * r;	
endfunction

//Calculates the empirical mean profit 
//x - list of performances
function r = mySum(x)
	y = thrownan(x);
	r = sum(y) ./ size(y,1);
endfunction

//calculates the optimal W, given instances and schedule
function W = calculateW(X,c)
		R = zeros(X);
		W = R; 
		ieee(2);
		for i=1:size(X,1), R(i,:) = X(i,:) ./ c'; end;
		for i=1:size(X,1), [m,n] = max(R(i,:)); W(i,n) = 1; end
endfunction

//Determines the best PPA from a series of runs, mean optimization
//trials - how many time the CMA should be executed
function [m, c] = iterativeCMA1(trials, X)
  m = -1000000;
  for i=1:trials do
    c2 = calculateCMA(X,1);
    m2 = mySum(simulate(X,c2));
    if m2 > m then
      m = m2;
      c = c2;
    end
  end
endfunction

//Determines the best PPA from a series of runs, limit optimization
//trials - how many time the CMA should be executed
function [m, c] = iterativeCMA2(trials, X)
  m = -1000000;
  for i=1:trials do
    c2 = calculateCMA(X,2);
    m2 = min(simulate(X,c2));
    if m2 > m then
      m = m2;
      c = c2;
    end
  end
endfunction


//Calculates the optimal schedule for a fixed classification
//t - type of the optimization, the same meaning as in calculateCMA
function c = calculateC(X,W,t)
	A = X.*W
	
	if t == 1 then
		a = sum(A, 'r');
		b = sum(W, 'r');
		ieee(2);		//eliminate errors from multiplication by zero
		a = naneliminate(a ./ b);
		b = normalize(b);	
		//use abs, because: sqrt(-0) = -0, 
		//and if a = -0 then 1/a=-inf, and that is wrong
		//it is caused by a scilab bug
		c =	abs(sqrt(-b .* a))'; 
	elseif t == 2 then
		a = min(A, 'r');
		c = a';
	end
	
	c = normalize(c);
endfunction

//Calculate the optimal c, W using the  
//t - type of optimization, 
//t = 1 means "mean optimization", t = 2 means "limit optimization".
function [c, W] = calculateCMA(X,t)
	c = normalize(rand(size(X,2),1));
	
	for i=1:20 do
		W = calculateW(X,c);
		cOld = c;
		c = calculateC(X,W,t);
		n = norm(c - cOld);
		//iterate only until the last element is reached
		if n == 0 then break; end;
	end
endfunction

//Column performance plot
function plotPerformance(p)
	p1 = tabul(p);
	p1(:,2) = cumsum(p1(:,2));
	p1(:,2) = p1(:,2) ./ size(p,1);
	plot2d(-p1(:,1), p1(:,2));
endfunction

//**** VOCMA algorithm


//VOCMA algorithm
//pc - the schedule
//p - the profit of the schedule
//t - the same as in calculateCMA
function [pc, p] = solveOptimal(X,t)
	m = size(X,1);
	n = size(X,2);
	k = size(R,2);
	R = buildStructure(X);
	p = -10000;
	pc = 0;
	
	o = zeros(k, 1);
	d = 0; vm = 0;
	while sum(abs(o - ones(k,1)*m)) > 0 do
		d = d+1;
		if modulo(d, 1000) == 0 then
			disp([d, vm, p, pc']);
		end;
		W = classifyVector(o, R, m, n);
		[W,v] = createW(W);
		if v then
			vm = vm + 1;
			c = calculateC(X,W,t);
			if t == 1 then
				e = mySum(simulate(X,c));
			else
				e = min(simulate(X,c));
			end;			
			if e > p then
				p = e;
				pc = c;
			end
			p = max(p, e);
		end
		o = createNextVector(o, m);
	end
endfunction

//Builds structure required to translate split points into 
//a classification
function R = buildStructure(X)
	n = size(X, 2);
	R = zeros(size(X,1), n * (n-1) /2);
	z = 0;
	for j=1:n do
		for k = (j+1):n do
			z = z + 1;
			a = X(:,j) ./ X(:,k);
			[b,c] = gsort(a, 'g', 'i');
			R(:,z) = c;
		end;
	end;
endfunction

//Creates a schedules for a set of split points
//the split point is defined in terms of the first algorithm
//h - the vector to classify
//R - the result of build structure - mapping to instances
//m - number of instances
//n - number of algorithms
function W = classifyVector(h, R, m, n)
	s = size(h,1);
	W = zeros(m, n);
	z = 0;
	for j=1:n do
		for k = (j+1):n do
			z = z + 1;
			W(R(1:h(z),z),j) = W(R(1:h(z),z),j) + 1/(n-1);
			W(R(h(z)+1:$,z),k) = W(R(h(z)+1:$,z),k) + 1/(n-1);
		end
	end
endfunction

//For a partially specified W, returns 0/1 W, if possible
//v specifies if each problem instance is solved
function [W, v] = createW(W)
	W = floor(W);
	v = (sum(W) == size(W,1));
endfunction

//Creates a next split point vector
//k is the limit of the highest number
function o = createNextVector(o, k)
	for i=size(o,1):-1:1 do
		if(o(i) >= k)
			o(i)=0;
		else
			o(i)=o(i)+1;
			break;
		end
	end		
endfunction

//*** helper functions

//Removes the rows that are not solved by either algorithm
function X = eliminateUnsolvable(X)
	r = find(max(X, 'c') <= -10000);
	X(r, :) = [];
endfunction

//Modifies the end value of the problem, 
//if the standard non finish profit is < 10000
function X = modifyEndValue(X, nval)
	X(find(X <= -10000)) = nval;
endfunction

//Normalizes the vector
function x = normalize(x)
	x = x ./ sum(x);
endfunction

//Replaces all occurrences of %nan and %inf by zero
function x = naneliminate(x)
	b = isnan(x);
	c = isinf(x);
	x(b) = 0;
	x(c) = 0;
endfunction

//Reverses the input matrix in the specified direction
function X = reverse(X, d)
	if d == 'r' then
		X = X($:-1:1,:)
	elseif d == 'c' then
		X = X(:,$:-1:1)
	end
endfunction

